/*
 * rs_shelling.c
 * Search shellable h-assignment by reverse search.
 *
 * by Masahiro Hachimori, Aug 2002.
 */

#include <stdio.h>
#include <stdlib.h>
#include "s_complex.h"

/*
 *
 *
 *
 */

/*
 * function to compute facet-ridge incidences.
 */
void get_incidence(fadj, radj, numf, numr, faces)
  int **fadj; /* list of incident ridges */
  int **radj; /* list of incident facets */
  int numf;
  int numr;
  Simplices *faces;
{
  Simplices *fhead, *rhead, *tmpf, *tmpr;
  int dim, i, j, num;

  dim = faces->vert[0];
  fhead = faces;
  rhead = faces;
  while (rhead->vert[0] == dim)
    rhead = rhead->next;

  i = 0;
  tmpf = fhead;
  while(tmpf->vert[0] == dim){
    fadj[i] = calloc(sizeof(int), (dim + 2));
    fadj[i][0] = dim + 1; /* the number of incident ridges is d+1 */
    tmpr = rhead;
    j = 0;
    num = 1;
    while(tmpr->vert[0] == dim-1){
      if (face_contained(tmpr->vert, tmpf->vert)){
        fadj[i][num] = j;
        num ++;
        if (num == dim + 2)
          break;
        }
      if (tmpr->next == NULL)
        break;
      else{
        tmpr = tmpr->next;
        j ++;}
      }
    i ++;
    tmpf = tmpf->next;
    }

  i = 0;
  tmpr = rhead;
  while(tmpr->vert[0] == dim-1){
    tmpf = fhead;
    num = 1; /* first just count incident facets */
    while(tmpf->vert[0] == dim){
      if (face_contained(tmpr->vert, tmpf->vert)){
        num ++;
        }
      tmpf = tmpf->next;
      }
    radj[i] = calloc(sizeof(int), num);
    radj[i][0] = num-1;
    num = 1; /* then write incident facets to the list */
    j = 0;
    tmpf = fhead;
    while(tmpf->vert[0] == dim){
      if (face_contained(tmpr->vert, tmpf->vert)){
        radj[i][num] = j;
        num ++;
        }
      tmpf = tmpf->next;
      j ++;
      }
    i ++;
    if (tmpr->next == NULL)
      break;
    else 
      tmpr = tmpr->next;
    }

  return;
}




/*
 * function of shellability checking.
 */
int rs_shellable(facets)
  Simplices *facets;
{
  int f_vec[MAXDIM], h_vec[MAXDIM], tmph_vec[MAXDIM];
  Simplices *faces, *tmpfaces;
  int dim, numf, numr, num_assigned;
  int **fadj, **radj, **fori, **rori, *hassign;
  int i, j, k, l, m, idx, b;
  int *fcancel, maxcancel, tmpmax, *rval, *fbd;
unsigned long nn = 0;

  /* preparation of data from facet list*/
  dim = facets->vert[0];
  if (dim == 0) return 1;

  faces = facets2faces(facets);

  get_fvector(faces, f_vec);
  f_to_h(f_vec, h_vec, facets->vert[0]);
  if (Mseq(h_vec, facets->vert[0]) == 0) return 0;
  for (i = 0; i <= dim +1; i ++)
    tmph_vec[i] = h_vec[i];

  tmpfaces = faces;
  numf = 0; /* count the number of facets */
  while(tmpfaces->vert[0] == dim){
    numf ++;
    tmpfaces = tmpfaces->next; }
  numr = 0;
  while(tmpfaces->vert[0] == dim-1){
    numr++; /* count the number of ridges */
    if (tmpfaces->next != NULL)
      tmpfaces = tmpfaces->next;
    else
      break;
    }
  fadj = calloc(sizeof(int*), numf);
  radj = calloc(sizeof(int*), numr);
  printf("%d %d\n", numf, numr);

  get_incidence(fadj, radj, numf, numr, faces);

  rval = calloc(sizeof(int), numr); /* valency of ridges */
  for (i = 0; i < numr; i ++)
    rval[i] = radj[i][0];

  fbd = calloc(sizeof(int), numf); /* dim+1 - # of boundary ridges */
  for (i = 0; i < numf; i ++)
    for (j = 1; j <= fadj[i][0]; j ++)
      if (rval[fadj[i][j]] != 1)
        fbd[i] ++;
    
  hassign = calloc(sizeof(int), numf);
  for (i = 0; i < numf; i ++) hassign[i] = -1; /* -1 means not assigned.*/

  fcancel = calloc(sizeof(int), numf); /* to remember cancellable facets */
  maxcancel = -1;


  /* Now start searching. */

  /* main loop */
  idx = 0; /* starting index of facet */
  num_assigned = 0;
  while(1){
/*
printf("%d: ", num_assigned);
for(i=0; i < numf; i ++)
if (hassign[i] == -1) printf("*"); else printf("%d", hassign[i]);
printf("\n");
printf("%d: ", num_assigned);
for(i=0; i < numf; i ++)
printf("%d", fcancel[i]);
printf("\n");
*/
    /* look for removable facets */
    for (i = idx; i < numf; i ++){
      if (hassign[i] == -1 && tmph_vec[fbd[i]] > 0){
        tmph_vec[fbd[i]] --;
        if (num_assigned == numf-1) { /* if the facet is the last one ... */
          hassign[i] = fbd[i];
          for(i=0; i < numf; i ++)
            if (hassign[i] == -1) printf("*"); else printf("%d", hassign[i]);
          printf("\n");
printf("%lu\n", nn);
          return(1);
          }
        if(tmph_vec[0] == 0){  /* 0 should be assigned only in the last */
          tmph_vec[fbd[i]] ++; /* if 0-assigning is possible, */
          i = numf;            /* stop further searching. */
          break;}
        idx = i; /* now idx shows the removable facet. */
        tmpmax = maxcancel;
        if (tmpmax > idx) {
          b = 0; /* first, check cancellability of tmpmax */
          for (j = 1; j <= fadj[tmpmax][0]; j ++) {
            if (rval[fadj[tmpmax][j]] == 1)
              for (k = 1; k <= radj[fadj[tmpmax][j]][0]; k ++) 
                if (radj[fadj[tmpmax][j]][k] == idx) {
                  b = 1;
                  break;}
              if (b == 1) break; }
          if (b == 1) { /* if not cancellable, check between tmpmax and idx */
            for (j = tmpmax-1; j > idx; j --) {
              if (fcancel[j] == 1) {
                b = 2;
                for (k = 1; k <= fadj[j][0]; k ++) { 
                  if (rval[fadj[j][k]] == 1)
                    for (l = 1; l <= radj[fadj[j][k]][0]; l ++)
                      if (radj[fadj[j][k]][l] == idx) {
                        b = 3;
                        break; }
                  if (b == 3) break;
                  }
                if (b == 2){
                  tmpmax = j;
                  break;
                  }
                }
              }
            if (j == idx) {
              tmpmax = j;
              }
            }
          }
        /* check whether the removing of the facet is approval or not. */
        if (tmpmax > idx) { /* if larger facet is cancellable, not remove.*/
          tmph_vec[fbd[i]] ++;
          continue; }
        else{
          /* renew cancellability */
          b = 0;
          for (j = 1; j <= fadj[i][0]; j ++)
            if (rval[fadj[i][j]] == 1)
              for (k = 1; k <= radj[fadj[i][j]][0]; k ++)
                if (fcancel[radj[fadj[i][j]][k]] >= 1){
                  fcancel[radj[fadj[i][j]][k]] ++; /* adjacent cancellable
                                                      becomes uncancellable.*/
                  if (radj[fadj[i][j]][k] == maxcancel)
                    b = 1;
                  }
          fcancel[i] = 1; /* removed one is cancellable */
          if (i > maxcancel) maxcancel = i;
          else
            if (b == 1)
              for (j = maxcancel-1; j >= i; j --)
                if (fcancel[j] == 1){
                  maxcancel = j;
                  break;}
          /* removing step */
          num_assigned ++;
          hassign[i] = fbd[i];
          for (j = 1; j <= fadj[i][0]; j ++) {
            rval[fadj[i][j]] --;
            if (rval[fadj[i][j]] == 1)
              for (k = 1; k <= radj[fadj[i][j]][0]; k ++)
                if (radj[fadj[i][j]][k] != i &&
                    hassign[radj[fadj[i][j]][k]] == -1) {
                  fbd[radj[fadj[i][j]][k]] --;
                  break;}
            }
          /* --- */
          idx = 0;
          break; 
          }
        }
      }

    if (i == numf) { /* if no removable facets are found ... */
      if (num_assigned == 0){
printf("%lu\n", nn);
        return(0);  /* if in the root, conclude no shellable h-assignment.*/
        }
      else {          /* otherwise go one step backwards. */
        j = maxcancel;
/* if (j == -1) {printf("error!\n");exit (-1);} */
              for (k = 1; k <= fadj[j][0]; k ++) /* resume cancellability */
                if (rval[fadj[j][k]] == 0)
                  for (l = 1; l <= radj[fadj[j][k]][0]; l ++)
                    if (fcancel[radj[fadj[j][k]][l]] >= 2){
                      fcancel[radj[fadj[j][k]][l]] --;
                      if (fcancel[radj[fadj[j][k]][l]] == 1 &&
                          maxcancel < radj[fadj[j][k]][l])
                        maxcancel = radj[fadj[j][k]][l]; }
              fcancel[j] = 0;
              if (maxcancel == j){
                for (k = j-1; k >= 0; k --)
                  if (fcancel[k] == 1){
                    maxcancel = k;
                    break;}
                if (k == -1) maxcancel = -1;}
              /* unremoving step */
              tmph_vec[hassign[j]] ++;
              hassign[j] = -1;
              num_assigned --;
              for (k = 1; k <= fadj[j][0]; k ++) {
                rval[fadj[j][k]] ++;
                if (rval[fadj[j][k]] == 2)
                  for (l = 1; l <= radj[fadj[j][k]][0]; l ++)
                  if (radj[fadj[j][k]][l] != j &&
                      hassign[radj[fadj[j][k]][l]] == -1) {
                    fbd[radj[fadj[j][k]][l]] ++;
                    break;}
                }
              /* --- */
              idx = j + 1;
nn++;
        /* for debugging only */
/*      if (j == -1) {
          printf("error!\n");
          exit(-1);
          } */
        }
      }
    }

}

